/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "mpid_bind.h"
#include "ibverbs_header.h"
#include "viutil.h"
#include "viapriv.h"
#include "viapacket.h"
#include "queue.h"
#include "viaparam.h"
#include "vbuf.h"
#include "ib_init.h"
#include "async_progress.h"

static int post_async_srq_recv(int count)
{
    vbuf *v;
    int i;
    struct ibv_recv_wr *bad_wr;

    for(i = 0; i < count; i++) {

        v = get_vbuf();

        assert(NULL != v);

        vbuf_init_recv(v, VBUF_BUFFER_SIZE);

        if(ibv_post_srq_recv(viadev.async_info.srq,
                    &v->desc.u.rr, &bad_wr)) {
            return 1;
        }
    }

    return 0;
}

static void viadev_ext_async_sendq_queue(
        viadev_connection_t * c, vbuf * v)
{
    v->desc.next = NULL;
    if (c->async_conn.ext_sendq_head == NULL) {
        c->async_conn.ext_sendq_head = v;
    } else {
        c->async_conn.ext_sendq_tail->desc.next = v;
    }
    c->async_conn.ext_sendq_tail = v;

    c->async_conn.ext_sendq_size++;
}

static inline int atomic_send_or_queue(
        viadev_connection_t *c, vbuf *v)
{
    struct ibv_send_wr *bad_wr;
    struct vbuf *qv;
    viadev_packet_header *p;

    ASYNC_THREAD_LOCK;

    if(IBV_WR_SEND == v->desc.u.sr.opcode) {
        /* Stamp the sequence number */
        p = (viadev_packet_header *) VBUF_BUFFER_START(v);
        p->id = c->next_packet_tosend++;
    }

    /* Flush existing sends on sendq */
    while (c->async_conn.send_wqes_avail && 
            c->async_conn.ext_sendq_head) {

        /* dequeue one vbuf */
        qv = c->async_conn.ext_sendq_head;

        c->async_conn.ext_sendq_head = qv->desc.next;

        if (qv == c->async_conn.ext_sendq_tail) {
            c->async_conn.ext_sendq_tail = NULL;
        }
        qv->desc.next = NULL;
        /* put the vbuf on the HCA's send Q */
        c->async_conn.send_wqes_avail--;
        c->async_conn.ext_sendq_size--;

        if(ibv_post_send(c->async_conn.qp, 
                    &(qv->desc.u.sr), &bad_wr)) {
            error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
        }
    }

    if(c->async_conn.send_wqes_avail <= 0) {
        viadev_ext_async_sendq_queue(c, v);
        ASYNC_THREAD_UNLOCK;
        return;
    }

    c->async_conn.send_wqes_avail--;

    if(ibv_post_send(c->async_conn.qp, &(v->desc.u.sr), &bad_wr)) {
        error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
    }

    if(viadev_use_apm && viadev_use_apm_test) {
        perform_manual_apm(c->vi);
    }

    ASYNC_THREAD_UNLOCK;

    return 0;
}

static inline void viadev_atomic_ext_async_sendq_send(
        viadev_connection_t * c)
{
    vbuf *v;
    struct ibv_send_wr *bad_wr;

    ASYNC_THREAD_LOCK;

    c->async_conn.send_wqes_avail++;

    while (c->async_conn.send_wqes_avail && 
            c->async_conn.ext_sendq_head) {

        /* dequeue one vbuf */
        v = c->async_conn.ext_sendq_head;

        c->async_conn.ext_sendq_head = v->desc.next;

        if (v == c->async_conn.ext_sendq_tail) {
            c->async_conn.ext_sendq_tail = NULL;
        }
        v->desc.next = NULL;
        /* put the vbuf on the HCA's send Q */
        c->async_conn.send_wqes_avail--;
        c->async_conn.ext_sendq_size--;

        if(ibv_post_send(c->async_conn.qp, 
                    &(v->desc.u.sr), &bad_wr)) {
            error_abort_all(IBV_RETURN_ERR,"Error posting send\n");
        }
    }

    ASYNC_THREAD_UNLOCK;
}

void viadev_post_send_async_qp(viadev_connection_t * c, vbuf * v)
{
    viadev_packet_header *p;
   
    p = (viadev_packet_header *) VBUF_BUFFER_START(v);

    v->grank = c->global_rank;

    v->desc.u.sr.send_flags = IBV_SEND_SIGNALED;

    atomic_send_or_queue(c, v);
}

static void viadev_post_rdmaread_async_qp(
        viadev_connection_t *c, vbuf *v)
{
    struct ibv_send_wr *bad_wr;

    D_PRINT("post_rdmaread: [vi %d] %s",
            c->global_rank, viadev_packet_to_string(c->vi, v));

    /* set id to be the global_rank */
    v->grank = c->global_rank;

    /* Don't try to inline this data, since it is large */
    v->desc.u.sr.send_flags = IBV_SEND_SIGNALED;

    atomic_send_or_queue(c, v);
}


static void viadev_async_recv_rget(MPIR_RHANDLE *rhandle)
{
    vbuf *v;
    viadev_packet_rget *rget_packet = NULL;

    viadev_connection_t *c = 
        (viadev_connection_t *) rhandle->connection;

    /* first, if receive buffer is NULL, must be a zero length
     * transfer, just use a dummy buffer location so that we have
     * something to transfer */
    if(NULL == rhandle->buf) {

        rhandle->buf = &nullrbuffer;
    }

    /* Remote address is NULL, cannot perform RDMA Read */
    if(NULL == rhandle->remote_address) {
        error_abort_all(IBV_STATUS_ERR,
                "RDMA read with null remote buffer "
                "not supported, rhandle : %p\n",
                rhandle);
    }

    rhandle->dreg_entry = NULL;

    viadev_register_recvbuf_if_possible(rhandle);

    if(NULL == rhandle->dreg_entry) {
        /* Sender specified RDMA read, but we are unable
         * to register the receive buffer, should have
         * an explicit message to sender to let him know of
         * this situation */
        error_abort_all(IBV_STATUS_ERR,
                "Registration of recvbuf failed - RDMA read failed\n");
    }

    /* Get vbuf for RDMA Read and issue it */
    v = get_vbuf();

    assert(v != NULL);

    rget_packet = (viadev_packet_rget *) VBUF_BUFFER_START(v);

    /* This field is set in the vbuf data area to allow us
     * to later find on RDMA read completion, which receive
     * request it corresponded to
     */

    rget_packet->rreq = rhandle;

    vbuf_init_rget(v, (void *)rhandle->buf, 
            ((struct dreg_entry *) rhandle->dreg_entry)->memhandle->rkey,
            (void*) rhandle->remote_address, 
            rhandle->remote_memhandle_rkey, 
            rhandle->len);

    viadev_post_rdmaread_async_qp(c, v);
}

static void incoming_async_rndz_start(vbuf *v, 
        viadev_connection_t *c,
        viadev_packet_rendezvous_start *header)
{
    MPIR_RHANDLE *rhandle;
    int found = 0;

    /* Only asynchronous RNDZ protocol messages
     * may arrive through this part of code */
    assert(VIADEV_PROTOCOL_ASYNC == 
            header->protocol);

    ASYNC_THREAD_LOCK;

    MPID_Msg_arrived(header->envelope.src_lrank,
                     header->envelope.tag,
                     header->envelope.context, &rhandle, &found);


    /* This information needs to be filled in whether a preposted 
     * receive was found or not.  */

    rhandle->connection = c;
    rhandle->send_id = header->sreq;
    rhandle->s.count = header->envelope.data_length;
    rhandle->vbufs_received = 0;
    rhandle->vbuf_head = NULL;
    rhandle->vbuf_tail = NULL;
    rhandle->replied = 0;
    rhandle->protocol = header->protocol;

    rhandle->remote_address = header->buffer_address;
    rhandle->remote_memhandle_rkey = header->memhandle_rkey;

    /* Unlock only after all rhandle fields have been
     * updated, since in the case where the message is
     * unexpected, the main thread may be pulling out
     * the rhandle as its being updated by the
     * asynchronous thread */

    if (!found) {
        /* recv not yet posted, set size of incoming data in
         * unexpected queue rhandle */
        rhandle->len = header->envelope.data_length;

        /* Unex msg to be handled just like RGET!! */
        rhandle->protocol = VIADEV_PROTOCOL_RGET;

        ASYNC_THREAD_UNLOCK;

    } else {
        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            error_abort_all(IBV_STATUS_ERR,
                            "message [%d, %d, %d] truncated. ask %d got %d"
                            " rhandle %p",
                            header->envelope.src_lrank,
                            header->envelope.tag,
                            header->envelope.context,
                            rhandle->len, 
                            header->envelope.data_length,
                            rhandle);
        }

        rhandle->len = header->envelope.data_length;

        /* packet now arrived, can no-longer cancel recv */
        rhandle->can_cancel = 0;

        ASYNC_THREAD_UNLOCK;

        viadev_async_recv_rget(rhandle);
    }
}

static void viadev_async_rget_finish(MPIR_RHANDLE * r)
{
    vbuf *v = get_vbuf();
    void *buffer = VBUF_BUFFER_START(v);
    viadev_packet_rget_finish *packet = buffer;
    viadev_connection_t *c = (viadev_connection_t *) (r->connection);

    /* fill in the packet */

    ASYNC_PACKET_SET_HEADER(packet, c, VIADEV_PACKET_RGET_FINISH);

    packet->sreq = REQ_TO_ID(r->send_id);

    v->shandle = r;

    /* prepare descriptor and post */
    vbuf_init_send(v, sizeof(viadev_packet_rget_finish));
    viadev_post_send_async_qp(c, v);
}


static void process_async_rdma_read_completion(vbuf *v)
{
    MPIR_RHANDLE *rhandle;
    viadev_connection_t *c = NULL;

    viadev_packet_rget *packet =
        (viadev_packet_rget *) VBUF_BUFFER_START(v);

    c = &(viadev.connections[v->grank]);

    rhandle = (MPIR_RHANDLE *) packet->rreq;

    /* Send the FIN msg */
    viadev_async_rget_finish(rhandle);

    v->shandle = NULL; /* For sanity!*/
}

static void incoming_async_rndz_rget_finish(vbuf *v,
        viadev_connection_t *c,
        viadev_packet_rget_finish *h)
{
    MPIR_SHANDLE *shandle;

    /* find the rhandle for this data */
    shandle = (MPIR_SHANDLE *) ID_TO_REQ(h->sreq);

    assert(VIADEV_PROTOCOL_ASYNC == shandle->protocol);
    SEND_COMPLETE(shandle);
}

static void process_async_send_completion(struct ibv_wc *wc)
{
    vbuf *v;
    viadev_connection_t *c;
    viadev_packet_header *header;

    v = (struct vbuf *) wc->wr_id;

    c = &(viadev.connections[v->grank]);

    viadev_atomic_ext_async_sendq_send(c);

#ifdef ADAPTIVE_RDMA_FAST_PATH
    if (v->padding == RGET_VBUF_FLAG) {
        process_async_rdma_read_completion(v);
        release_vbuf(v);
        return;
    }
#else
    if (v->desc.u.sr.opcode == IBV_WR_RDMA_READ) {
        process_async_rdma_read_completion(v);
        release_vbuf(v);
        return;
    }
#endif                          /* ADAPTIVE_RDMA_FAST_PATH */

    header = (viadev_packet_header *) VBUF_BUFFER_START(v);

    switch (header->type) {
        case VIADEV_PACKET_RENDEZVOUS_START:
#ifdef ADAPTIVE_RDMA_FAST_PATH
            if (v->padding == NORMAL_VBUF_FLAG) {
                release_vbuf(v);
            } else {
                v->padding = FREE_FLAG;
                v->len = 0;
            }
#else
            release_vbuf(v);
#endif
            break;
        case VIADEV_PACKET_RGET_FINISH:
            {
                MPIR_RHANDLE *r = NULL;
                r = (MPIR_RHANDLE *) v->shandle;

                if (r != NULL) {
                    RECV_COMPLETE(r);
                }
            }
#ifdef ADAPTIVE_RDMA_FAST_PATH
            if (v->padding == NORMAL_VBUF_FLAG) {
                release_vbuf(v);
            } else {
                v->padding = FREE_FLAG;
                v->len = 0;
            }
#else
            release_vbuf(v);
#endif
            break;
        default:
            error_abort_all(GEN_EXIT_ERR,
                    "Unknown send completion type (%d) in "
                    "ASYNC thread\n", header->type);
    }
}

static void process_async_recv_completion(struct ibv_wc *wc)
{
    vbuf *v;
    viadev_connection_t *c;
    int done = 0;
    viadev_packet_header *header;

    v = (struct vbuf *) wc->wr_id;

    header = (viadev_packet_header *)
        VBUF_BUFFER_START(v);

    /* Message over here can only be from SRQ! */
    c = &(viadev.connections[header->src_rank]);

    v->grank = header->src_rank; /* For all the other layers of code */

    /* Check message sequence id for sanity */
    if(header->id != c->next_packet_expected) {

        /* Main thread will process this message */

        VBUF_QUEUE_ENQUE((&viadev.async_info.vbufq), v);
        post_async_srq_recv(1);

        return;
    }

    ASYNC_THREAD_LOCK;

    c->next_packet_expected++;

    ASYNC_THREAD_UNLOCK;

    /* Replenish SRQ */
    post_async_srq_recv(1);

    switch(header->type) {
        case VIADEV_PACKET_RENDEZVOUS_START:
            incoming_async_rndz_start(v, c,
                    (viadev_packet_rendezvous_start *) header);
#ifdef ADAPTIVE_RDMA_FAST_PATH
            if (v->padding == NORMAL_VBUF_FLAG) {
                release_vbuf(v);
            } else {
                assert(0);
                release_recv_rdma(c, v);
            }
#else
            release_vbuf(v);
#endif
            break;
        case VIADEV_PACKET_RGET_FINISH:
            incoming_async_rndz_rget_finish(v, c,
                    (viadev_packet_rget_finish *) header);
#ifdef ADAPTIVE_RDMA_FAST_PATH
            if (v->padding == NORMAL_VBUF_FLAG) {
                release_vbuf(v);
            } else {
                assert(0);
                release_recv_rdma(c, v);
            }
#else
            release_vbuf(v);
#endif
            break;
        default:
            error_abort_all(GEN_ASSERT_ERR, 
                    "Unknown Recv Completion type (%d) "
                    "in ASYNC thread\n", header->type);
    }

}

#define PROCESS_ASYNC_CQ_EVENT {                           \
    vbuf_addr = (void *) ((aint_t) sc.wr_id);              \
    if (sc.status != IBV_WC_SUCCESS) {                     \
        error_abort_all(((vbuf *) vbuf_addr)->grank,        \
                "[%s:%d] Got completion with error %s, "    \
                "code=%d, dest rank=%d\n",                  \
                viadev.my_name, viadev.me,                  \
                wc_code_to_str(sc.status), sc.status,       \
                ((vbuf *) vbuf_addr)->grank);               \
    }                                                       \
    if (sc.opcode == IBV_WC_SEND ||                         \
            sc.opcode == IBV_WC_RDMA_WRITE ||               \
            sc.opcode == IBV_WC_RDMA_READ) {                \
        /* Send Completion */                               \
        process_async_send_completion(&sc);                 \
    } else {                                                \
        /* Recv Completion */                               \
        process_async_recv_completion(&sc);                 \
    }                                                       \
}

static void async_completion_thread(void * ctx)
{
    int ret;
    struct ibv_comp_channel *ev_ch;
    struct ibv_cq *ev_cq;
    void *ev_ctx;
    int ne;
    void *vbuf_addr;
    struct ibv_wc sc;
    unsigned long long nspin = 0;

    /* This thread should be in a cancel enabled state */
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

    ev_ch = viadev.async_info.comp_channel;

    while(1) {
        pthread_testcancel(); 
        pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
        do {
            ret = ibv_get_cq_event(ev_ch, &ev_cq, &ev_ctx);

            if (ret && errno != EINTR) {
                error_abort_all(IBV_RETURN_ERR,
                        "Failed to get cq event: %d\n", ret);
            }

        } while (ret && errno == EINTR);
        pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

        if (ev_cq != viadev.async_info.cq) {
            error_abort_all(GEN_ASSERT_ERR, "Event in unknown CQ\n");
        }

        ibv_ack_cq_events(viadev.async_info.cq, 1);

        pthread_testcancel(); 


#if 0
        fprintf(stderr,"[%d] Woke up!\n", viadev.me);
#endif
        pthread_testcancel(); 

        while (1) {
            ne = ibv_poll_cq(viadev.async_info.cq, 1, &sc);
            pthread_testcancel(); 

            if (ne > 0) {
                PROCESS_ASYNC_CQ_EVENT;

                /* Found something, erase poll counter */
                nspin = 0;
            } else if (ne < 0) {
                error_abort_all(IBV_RETURN_ERR, "[%s:%d] Got error polling CQ\n", 
                        viadev.my_name, viadev.me);
            } else {
                nspin++;

                if(nspin == viadev_max_async_spin_count) {
                    break;
                }
            }
        }

        if (ibv_req_notify_cq(viadev.async_info.cq, 0)) {
            error_abort_all(IBV_RETURN_ERR,
                    "Couldn't request for CQ notification\n");
        }

        while (1) {
            ne = ibv_poll_cq(viadev.async_info.cq, 1, &sc);
            pthread_testcancel(); 

            if (ne > 0) {
                PROCESS_ASYNC_CQ_EVENT;
                /* Found something, erase poll counter */
                nspin = 0;
            } else if (ne < 0) {
                error_abort_all(IBV_RETURN_ERR, "[%s:%d] Got error polling CQ\n", 
                        viadev.my_name, viadev.me);
            } else {
                break;
            }
        }
    }
}

int viadev_init_async_progress()
{
    struct ibv_cq *cq = NULL;
    struct ibv_comp_channel *ch = NULL;
    struct ibv_srq_init_attr srq_init_attr;
    struct ibv_srq *srq = NULL;

    memset(&viadev.async_info, 0, sizeof(async_prog_info_t));

    if(pthread_spin_init(&viadev.async_info.lock, 
                PTHREAD_PROCESS_PRIVATE)) {
        return 1;
    }

    if(pthread_spin_init(&viadev.async_info.vbufq.lock, 
                PTHREAD_PROCESS_PRIVATE)) {
        return 1;
    }


    ch = ibv_create_comp_channel(viadev.context);

    if (!ch) {
        return 1;
    }

    cq = ibv_create_cq(viadev.context, viadev_cq_size,
            NULL, ch, 0);

    if (!cq) {
        return 1;
    }

    viadev.async_info.cq = cq;
    viadev.async_info.comp_channel = ch;

    if(ibv_req_notify_cq(cq, 0)) {
        return 1;
    }

    memset(&srq_init_attr, 0, sizeof(srq_init_attr));

    srq_init_attr.srq_context = viadev.context;
    srq_init_attr.attr.max_wr = viadev_async_srq_size;
    srq_init_attr.attr.max_sge = 1;
    /* The limit value should be ignored during SRQ create */
    srq_init_attr.attr.srq_limit = viadev_srq_limit;

    srq = ibv_create_srq(viadev.ptag, &srq_init_attr);

    if (!srq) {
        error_abort_all(IBV_RETURN_ERR, "Error creating SRQ\n");
    }

    viadev.async_info.srq = srq;

    if(post_async_srq_recv(viadev_async_srq_pool_size)) {
        return 1;
    };

    if(pthread_create(&viadev.async_info.completion_thread,
            NULL, (void *) async_completion_thread,
            (void *) viadev.context)) {
        return 1;
    }

    return 0;
}

int viadev_finalize_async_progress()
{
    async_prog_conn_t *async_conn = NULL;

    while(1) {

        REMOVE_ASYNC_CONN((viadev.async_info), &async_conn);

        if(NULL == async_conn) {
            break;
        }
        if(ibv_destroy_qp(async_conn->qp)) {
            error_abort_all(GEN_ASSERT_ERR,
                    "Failed to destroy QP\n");
        }
    }

    /* Cancel thread if active */
    if (pthread_cancel(viadev.async_info.completion_thread)) {
        error_abort_all(GEN_ASSERT_ERR,
                "Failed to cancel async completion thread\n");
    }

    if (pthread_join(viadev.async_info.completion_thread, NULL)) {
        error_abort_all(GEN_ASSERT_ERR,
                "Failed to join async completion thread\n");
    }

    if (ibv_destroy_srq(viadev.async_info.srq)) {
        error_abort_all(IBV_RETURN_ERR,
                "Error destroying ASYNC SRQ\n");
    }

    if (ibv_destroy_cq(viadev.async_info.cq)) {
        error_abort_all(IBV_RETURN_ERR,
                "Error destroying ASYNC CQ\n");
    }

    if (ibv_destroy_comp_channel
            (viadev.async_info.comp_channel)) {
        error_abort_all(IBV_RETURN_ERR,
                "Error destroying ASYNC Comp. Channel\n");
    }

    if (pthread_spin_destroy(&viadev.async_info.lock)) {
        error_abort_all(IBV_RETURN_ERR,
                "Error destroying Pthread Spin lock\n");
    }

    return 0;
}

int viadev_create_async_qp(viadev_connection_t *c)
{
    struct ibv_qp *qp = NULL;
    struct ibv_qp_init_attr init_attr;
    struct ibv_qp_attr attr;

    memset(&init_attr, 0, sizeof(struct ibv_qp_init_attr));

    init_attr.send_cq = viadev.async_info.cq;
    init_attr.recv_cq = viadev.async_info.cq;
    init_attr.srq = viadev.async_info.srq;
    init_attr.cap.max_send_wr = viadev_async_sq_size;
    init_attr.cap.max_send_sge = 1;
    init_attr.cap.max_recv_sge = 1;
    init_attr.cap.max_inline_data = 0;
    init_attr.qp_type = IBV_QPT_RC;

    qp = ibv_create_qp(viadev.ptag, &init_attr);

    if (!qp) {
        return 1;
    }

    c->async_conn.qp = qp;
    c->async_conn.send_wqes_avail = viadev_async_sq_size;
    c->async_conn.ext_sendq_head = NULL;
    c->async_conn.ext_sendq_size = 0;
    c->async_conn.next_packet_tosend = 0;
    c->async_conn.next_packet_expected = 0;

    memset(&attr, 0, sizeof(struct ibv_qp_attr));
    attr.qp_state = IBV_QPS_INIT;
    set_pkey_index(&attr.pkey_index, viadev_default_port);
    attr.port_num = viadev_default_port;
    attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE |
        IBV_ACCESS_REMOTE_READ;
    cm_ib_qp_attr.rc_qp_mask_to_init = IBV_QP_STATE |
        IBV_QP_PKEY_INDEX | IBV_QP_PORT | IBV_QP_ACCESS_FLAGS;

    if(ibv_modify_qp(qp, &attr,
                      IBV_QP_STATE | 
                      IBV_QP_PKEY_INDEX | IBV_QP_PORT | 
                      IBV_QP_ACCESS_FLAGS)) {
        return 1;
    }

    ADD_ASYNC_CONN( (viadev.async_info), &(c->async_conn) );

    return 0;
}

int viadev_transition_async_qp(viadev_connection_t *c,
        uint32_t dest_qpn, uint16_t dest_lid)
{
    struct ibv_qp_attr attr;
    struct ibv_qp *qp;

    qp = c->async_conn.qp;

    /* <--------- RTR ---------> */
    memset(&attr, 0, sizeof(struct ibv_qp_attr));
    attr.qp_state = IBV_QPS_RTR;
    attr.path_mtu = viadev_default_mtu;
    attr.max_dest_rd_atomic = viadev_default_qp_ous_rd_atom;
    attr.min_rnr_timer = viadev_default_min_rnr_timer;
    attr.ah_attr.sl = viadev_default_service_level;
    attr.ah_attr.src_path_bits = viadev_default_src_path_bits;
    attr.ah_attr.port_num = viadev_default_port;
    attr.ah_attr.static_rate = viadev_default_static_rate;
    attr.ah_attr.dlid = dest_lid;
    attr.dest_qp_num = dest_qpn;

    if(ibv_modify_qp(qp, &attr,
                IBV_QP_STATE | IBV_QP_AV | IBV_QP_PATH_MTU |
                IBV_QP_DEST_QPN | IBV_QP_RQ_PSN |
                IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER)) {
        return 1;
    }

    /* <--------- RTS ---------> */

    memset(&attr, 0, sizeof(struct ibv_qp_attr));
    attr.qp_state = IBV_QPS_RTS;
    attr.timeout = viadev_default_time_out;
    attr.retry_cnt = viadev_default_retry_count;
    attr.rnr_retry = viadev_default_rnr_retry;
    attr.sq_psn = viadev_default_psn;
    attr.max_rd_atomic = viadev_default_qp_ous_rd_atom;

    if(ibv_modify_qp(qp, &attr, IBV_QP_STATE | IBV_QP_TIMEOUT | 
                IBV_QP_RETRY_CNT | IBV_QP_RNR_RETRY | 
                IBV_QP_SQ_PSN | IBV_QP_MAX_QP_RD_ATOMIC)) {
        return 1;
    }

    if(viadev_use_apm) {
        reload_alternate_path(qp);
    }

    return 0;
}
