/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include "mpid_bind.h"
#include "ibverbs_header.h"
#include "mv_priv.h"
#include "mv_packet.h"
#include "queue.h"
#include "mv_param.h"
#include "mpid_smpi.h"
#include "mv_buf.h"
#include "mv_inline.h"

#define ACK_CREDIT_CHECK(_c, _v) {                                     \
    if(MVDEV_TRANSPORT_UD == _v->transport) {                          \
        if(++(_c->ud_remote_credit) > mvparams.send_ackafter) {        \
            mvdev_explicit_ack(_c->global_rank);                       \
        }                                                              \
    }                                                                  \
}

#define SET_BYTE_LEN(_v, _wc) {                                                  \
    if(MVDEV_TRANSPORT_UD == _v->transport) {                                    \
        _v->byte_len_data = _v->byte_len = _wc.byte_len - 40 - MV_HEADER_OFFSET; \
    } else {                                                                     \
        _v->byte_len_data = _v->byte_len = _wc.byte_len - MV_HEADER_OFFSET;      \
    }                                                                            \
}

static inline int mvdev_process_send(mv_sdescriptor *d, mv_sbuf * v) {
    mv_qp *qp = (mv_qp *) d->qp;

    qp->send_wqes_avail++;
    assert(qp->send_wqes_avail <= qp->send_wqes_total);
    if(NULL != qp->ext_sendq_head && qp->send_wqes_avail > 0) {
        mvdev_ext_sendq_send(qp);
    } 

    v->left_to_send--;

    if(v->left_to_send == 0) {
        v->in_progress = 0;

        /* immediately free packets that are unacknowledged */
        if(v->seqnum == 0) { 
            mvdev_packet_header *p = (mvdev_packet_header *) v->header_ptr;
            assert(p->type == MVDEV_PACKET_ACK || v->in_sendwin == 0);
            release_mv_sbuf(v);
        }
    } 

    return 0;
}


static inline int mvdev_process_recv(mv_rbuf * v)
{
    mvdev_connection_t *c = &(mvdev.connections[v->rank]);

    /* this took something from the buffer pool */
    /* TODO: keep track of what pool/QP it came from... */
    mvdev.posted_bufs--;
    CHECK_RECV_BUFFERS();

    if(v->has_header) {
        D_PRINT("recv'd packet has header\n");

        mvdev_packet_header *p = (mvdev_packet_header *) v->header_ptr;
        mvdev_process_ack(c, p->last_recv);

        switch(p->type) {
            case MVDEV_PACKET_EAGER_START:
                MV_SBUF_SET_DATA(v, mvdev_packet_eager_start);
                break;
            case MVDEV_PACKET_EAGER_NEXT:
                MV_SBUF_SET_DATA(v, mvdev_packet_eager_next);
                break;
            case MVDEV_PACKET_R3_DATA:
                MV_SBUF_SET_DATA(v, mvdev_packet_r3_data);
                break;
        }

        switch(p->type) {
            case MVDEV_PACKET_ACK:
                assert(v->seqnum == 0);
                release_mv_rbuf(v);
                break;  
            default:
                assert(v->seqnum != 0);
                ACK_CREDIT_CHECK(c, v);
                mvdev_place_recvwin(c, v);
        }       
    } else {
        ACK_CREDIT_CHECK(c, v);
        mvdev_place_recvwin(c, v);
    }
    return 0;
}

void mvdev_process_send_afterack(mv_sbuf * v) {
    mvdev_packet_header *p;

    p = (mvdev_packet_header *) v->header_ptr;

    switch(p->type) {
        case MVDEV_PACKET_EAGER_START:
        case MVDEV_PACKET_EAGER_NEXT:
        case MVDEV_PACKET_RENDEZVOUS_START:
        case MVDEV_PACKET_RENDEZVOUS_REPLY:
        case MVDEV_PACKET_UD_ZCOPY_FINISH:
        case MVDEV_PACKET_UD_ZCOPY_ACK:
            goto release_vbuf;
            break;  
        case MVDEV_PACKET_R3_DATA:
            {
                mvdev_connection_t *c = &(mvdev.connections[v->rank]);
                c->r3_segments -= v->segments;
                if(v->in_progress) {
                    v->seqnum = 0;
                } else {
                    release_mv_sbuf(v);
                }

                if(c->r3_segments < mvparams.r3_segments_max && 
                        c->blocked_count > 0) {
                    PUSH_FLOWLIST(c);
                }
                return;
                goto release_vbuf;
                break;  
            }
        default:
            error_abort_all(IBV_STATUS_ERR, "Unknown packet type %d "
                    " id: %d "
                    "in mvdev_process_send_afterack", p->type, v->seqnum); 
    }

release_vbuf:
    if(v->in_progress) {
        v->seqnum = 0;
        I_PRINT("here!\n");
    } else {
        release_mv_sbuf(v);
    }
}


int MPID_DeviceCheck(MPID_BLOCKING_TYPE blocking)
{
    int ne = 0, progress = 0, count = 0;
    int empty = 1, i;
    struct ibv_wc wc;
    static long last_check = 0;
    static int recv = 0;

    do {    

        /* Poll SMP */
#ifdef _SMP_
        if (!disable_shared_mem && SMP_INIT && 
                MPI_SUCCESS == MPID_SMP_Check_incoming()) {
            progress = 1;
        }
#endif
        /* Poll the CQ */

        for(i = 0; i < mvdev.num_cqs; i++) {
            ne = ibv_poll_cq(mvdev.cq[i], 1, &wc);
            if(ne < 0) {
                error_abort_all(IBV_RETURN_ERR, "Error polling CQ\n");
            } else if (ne > 1) {
                error_abort_all(IBV_RETURN_ERR, "Asked only one completion, "
                        "got more");
            } else if (1 == ne) {
                if (wc.status != IBV_WC_SUCCESS) {
                    error_abort_all(IBV_STATUS_ERR, "got completion with "
                            "error. code: %d, wr_id: %lu\n", wc.status, wc.wr_id);
                }       

                switch(wc.opcode) {
                    case IBV_WC_RDMA_WRITE:
                    case IBV_WR_RDMA_READ:
                    case IBV_WC_SEND:
                        {
                            mv_sdescriptor *d = (mv_sdescriptor *) wc.wr_id;
                            D_PRINT("Got a SEND COMP: sbuf: %p, desc: %p\n", 
                                    d->parent, d);
                            mvdev_process_send(d, d->parent);
                            break;
                        }
                    case IBV_WC_RECV:
                        {
                            mv_rdescriptor *d = (mv_rdescriptor *) wc.wr_id;;
                            mv_rbuf *v = d->parent;

                            D_PRINT("Got a RECV: rbuf: %p\n", v);

                            if(mvparams.use_headers) {
                                mvdev_packet_mheader *mh = (mvdev_packet_mheader *) v->mheader_ptr;
                                v->has_header = mh->has_header;
                                v->rank       = mh->src_rank;
                                v->seqnum     = mh->seqnum;
                                D_PRINT("has_header: %d, rank: %d, "
                                        "seqnum: %d\n", v->has_header,
                                        v->rank, v->seqnum);
                            } else {
                                READ_THS_TUPLE(wc.imm_data, &(v->has_header), 
                                        &(v->rank), &(v->seqnum));
                            }

                            v->data_ptr = v->header_ptr;

                            SET_BYTE_LEN(v, wc);

                            mvdev_process_recv(v);
                            recv++;
                            break;
                        }
                    default:
                        error_abort_all(IBV_RETURN_ERR, "Invalid opcode\n");
                }

                progress = 1;
                empty = 0;
            } else {
                empty = 1;
                count++;

                if(count % 550 == 0) {
                    if(get_time_us() - last_check > mvparams.progress_timeout) {
                        last_check = get_time_us();
                        mvdev_unackq_traverse();
                        SEND_ACKS();
                        recv = 0;
                    }
                    else if(recv > 10) {
                        SEND_ACKS();
                        recv = 0;
                    }
                    CHECK_RECV_BUFFERS_LAZY();
                }
            }
        }

        if(flowlist) {
            process_flowlist();
        }

    } while ((blocking && !progress) || !empty);

    if(recv > mvparams.send_ackafter_progress || 
            (get_time_us() - last_check > mvparams.progress_timeout)) {
        mvdev_unackq_traverse();
        SEND_ACKS();
        last_check = get_time_us();
        recv = 0;
    }

    return progress ? MPI_SUCCESS : -1;
}

