
/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "mv_param.h"
#include "mv_buf.h"
#include "mv_priv.h"
#include "mv_packet.h"
#include "ibverbs_const.h"
#include "mpid_smpi.h"

mvdev_param_info mvparams;

#define DEFAULT_HCA_ID      (0)

#ifdef _SMP_
    int disable_shared_mem = 0;
    int smp_num_send_buffer = SMP_NUM_SEND_BUFFER;
    int smp_batch_size = SMP_BATCH_SIZE;
#ifdef _AFFINITY_
    unsigned int viadev_enable_affinity = 1;
    char *cpu_mapping = NULL;
#endif
#endif

void MV_Init_Params(int num_proc, int me) 
{
    char *value;

    D_PRINT("Entering MV_Init_Params\n");

    mvparams.rendezvous_threshold = 16384;

    mvparams.use_lmc = 1;
    mvparams.max_lmc_total = 1;
    mvparams.max_sl = 1;

    /* ZCOPY TRANSFER ----------------------------------------- */

    /* Use the zero-copy UD transfer method */
    mvparams.use_zcopy = 1;

    /* Message sizes >= threshold go over zcopy */
    mvparams.zcopy_threshold = 32768;

    /* QPs to be in rndv pool */
    mvparams.rndv_qps = 64;

    /* RELIABILITY TUNING ------------------------------------- */

    /* Time (usec) until ACK status is checked (and ACKs are sent) */
    mvparams.progress_timeout = 1400000;

    /* Time (usec) until a message is resent */
    mvparams.retry_usec = 20000000;

    /* Maximum number of segments outstanding (waiting for ACK) 
     * to any given host 
     */
    mvparams.send_window_segments_max = 500;

    /* Maximum number of out-of-order messages that will be buffered */
    mvparams.recv_window_size_max = 2501;

    /* Maximum number of messages received until an ACK is automatically
     * generated (not waiting for mvparams.progress_timeout */
    mvparams.send_ackafter = 50;
    mvparams.send_ackafter_progress = 10;

    /* maximum number of times to retry sending the same message */
    mvparams.max_retry_count = 50;


    /* QP PARAMETERS ------------------------------------------ */

    mvparams.cq_size = 40000;
    mvparams.sq_size = 2000;
    mvparams.rq_size = 5000;
    mvparams.psn = 0;

    mvparams.max_ud_inline = 60;

    /* RECEIVE BUFFERS ---------------------------------------- */

    /* Total number of recieve buffers to be posted */
    mvparams.recvq_size = mvparams.rq_size;


    /* MEMORY REGISTRATION ------------------------------------ */

    /* Use the registration cache */
    mvparams.use_dreg_cache = 1;

    /* number of entries */
    mvparams.ndreg_entries = 1000;
    mvparams.dreg_cache_limit = 0;

    /* OTHER PARAMETERS --------------------------------------- */

    mvparams.r3_segments_max = 300;

    mvparams.default_port = 1;

    /* 0: Utilize the immediate data for sequence numbers, on normal
     *    messages 
     * 1: Do not use the immediate data. Required for using more than 128K
     *    MPI ranks
     */
    mvparams.use_headers = 0;

    /* How many QPs are used to send messages on (round-robin) */
    mvparams.num_qps = 4;

    /* Maximum message transfer size for UD.
     *   -1: maximum that is active
     *    *: otherwise use the value specified
     */
    mvparams.mtu = -1;

    /* Get the appropriate IB device */
    strncpy(mvparams.device_name, MVDEV_INVALID_DEVICE, 32);
    if ((value = getenv("MV_DEVICE")) != NULL) {
        strncpy(mvparams.device_name, value, 32);
    }
    if ((value = getenv("MV_USE_LMC")) != NULL) {
       	mvparams.use_lmc = atoi(value);
    }
    if ((value = getenv("MV_MAX_VL")) != NULL) {
       	mvparams.max_sl = atoi(value);
        if(mvparams.max_sl <= 0) { mvparams.max_sl = 1; }
        else if(mvparams.max_sl > 14) { mvparams.max_sl = 14; }
    }
    if ((value = getenv("MV_PROGRESS_TIMEOUT")) != NULL) {
       	mvparams.progress_timeout = atoi(value);
    }
    if ((value = getenv("MV_RETRY_TIMEOUT")) != NULL) {
       	mvparams.retry_usec = atoi(value);
    }
    if ((value = getenv("MV_MAX_RETRY_COUNT")) != NULL) {
       	mvparams.max_retry_count = atoi(value);
    }
    if ((value = getenv("MV_ACK_AFTER_RECV")) != NULL) {
        mvparams.send_ackafter = atoi(value);
    }
    if ((value = getenv("MV_ACK_AFTER_PROGRESS")) != NULL) {
        mvparams.send_ackafter_progress = atoi(value);
    }
    if ((value = getenv("MV_USE_UD_ZCOPY")) != NULL) {
       	mvparams.use_zcopy = atoi(value);
    }
    if ((value = getenv("MV_UD_ZCOPY_QPS")) != NULL) {
       	mvparams.rndv_qps = atoi(value);
    }
    if ((value = getenv("MV_UD_ZCOPY_THRESHOLD")) != NULL) {
       	mvparams.zcopy_threshold = atoi(value);
    }
    if ((value = getenv("MV_USE_REG_CACHE")) != NULL) {
       	mvparams.use_dreg_cache = atoi(value);
    }
    if ((value = getenv("MV_USE_HEADERS")) != NULL) {
       	mvparams.use_headers = atoi(value);
    }
    if ((value = getenv("MV_NUM_UD_QPS")) != NULL) {
       	mvparams.num_qps = atoi(value);
    }
    if ((value = getenv("MV_UD_SQ_SIZE")) != NULL) {
       	mvparams.sq_size = atoi(value);
    }
    if ((value = getenv("MV_UD_RQ_SIZE")) != NULL) {
       	mvparams.rq_size = atoi(value);
        mvparams.recvq_size = atoi(value);
    }
    if ((value = getenv("MV_UD_RQ_SIZE_POSTED")) != NULL) {
       	mvparams.recvq_size = atoi(value);
    }
    if ((value = getenv("MV_UD_CQ_SIZE")) != NULL) {
       	mvparams.cq_size = atoi(value);
    }
    if ((value = getenv("MV_PSN")) != NULL) {
       	mvparams.psn = atoi(value);
    }
    if ((value = getenv("MV_RNDV_THRESHOLD")) != NULL) {
        mvparams.rendezvous_threshold = atoi(value);
        if(mvparams.rendezvous_threshold <= 0) {
            mvparams.rendezvous_threshold = 1;
        }
    }
    if ((value = getenv("MV_MTU")) != NULL) {
            mvparams.mtu = atoi(value);
    }

#ifdef _SMP_
    /* Run parameter to disable shared memory communication */
    if ( (value = getenv("MV_USE_SHARED_MEM")) != NULL ) {
        disable_shared_mem = !atoi(value);
    }
    /* Should set in KBytes */
    if ( (value = getenv("MV_SMP_EAGERSIZE")) != NULL ) {
        smp_eagersize = atoi(value);
    }
    /* Should set in KBytes */
    if ( (value = getenv("MV_SMPI_LENGTH_QUEUE")) != NULL ) {
        smpi_length_queue = atoi(value);
    }
    if ( (value = getenv("MV_SMP_NUM_SEND_BUFFER")) != NULL ) {
        smp_num_send_buffer = atoi(value);
    }
    if ( (value = getenv("MV_SMP_BATCH_SIZE")) != NULL ) {
        smp_batch_size = atoi(value);
    }

#ifdef _AFFINITY_
    if ((value = getenv("MV_ENABLE_AFFINITY")) != NULL) { 
        viadev_enable_affinity = atoi(value);
    }
   
    /* The new alias will override the old 
     * variable, if defined */

    if ((value = getenv("MV_USE_AFFINITY")) != NULL) {
        viadev_enable_affinity = atoi(value);
    }

    if ((value = getenv("VIADEV_CPU_MAPPING")) != NULL) {
        cpu_mapping = (char *)malloc(sizeof(char) * strlen(value) + 1);
        strcpy(cpu_mapping, value);
    }
#endif
#endif

    /* Re-post buffers when the number of buffers
     * posted decreases below this threshold 
     *   - mvparams.credit_preserve: hard limit
     *   - mvparams.credit_preserve_lazy: only if no messages
     *        have been coming in recently
     */
    mvparams.credit_preserve = mvparams.recvq_size - 275;
    mvparams.credit_preserve_lazy = mvparams.recvq_size - 50;

    D_PRINT("Exiting MV_Init_Params\n");
}

