/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef _MV_PRIV_H
#define _MV_PRIV_H

#include <pthread.h>
#include "mv_param.h"
#include "mv_packet.h"
#include "ibverbs_header.h"
#include "mv_buf.h"
#include "req.h"
#include "mpid.h"
#include "ibverbs_header.h"
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include "mv_priv.h"
#include "process/pmgr_collective_client.h"

#define GEN_EXIT_ERR     -1     /* general error which forces us to abort */
#define GEN_ASSERT_ERR   -2     /* general assert error */
#define IBV_RETURN_ERR   -3     /* ibverbs funtion return error */
#define IBV_STATUS_ERR   -4     /* ibverbs funtion status error */

#define error_abort_all(code, args...)  {                          \
        pmgr_abort(code, ##args);                                  \
        exit(code);                                                \
}

#define COLLECT_MPI_MSG_INFO 

#define MAX_SEQ_NUM 8191
#define WINDOW_ABS_MAX  100

#define UD_HOSTSEQ_ALL  0xffffffff
#define UD_SEQ_BITS 13
#define UD_HOST_BITS 18
#define UD_TYPE_BITS 1

#define UD_TYPE_MASK ~(UD_HOSTSEQ_ALL << UD_TYPE_BITS)
#define UD_SEQ_MASK (UD_HOSTSEQ_ALL << (32 - UD_SEQ_BITS))
#define UD_HOST_MASK ~(UD_TYPE_MASK | UD_SEQ_MASK)

#define CREATE_THS_TUPLE(_type, _host, _seq)     \
(((_type) & UD_TYPE_MASK) |                   \
((_host << UD_TYPE_BITS) & UD_HOST_MASK) |    \
((_seq << (UD_TYPE_BITS + UD_HOST_BITS)) & UD_SEQ_MASK))

#define READ_THS_TUPLE(_tuple, _type, _host, _seq)     \
{ \
    *_type = (_tuple & UD_TYPE_MASK); \
    *_host = (_tuple & UD_HOST_MASK) >> UD_TYPE_BITS; \
    *_seq  = (_tuple & UD_SEQ_MASK) >> (UD_TYPE_BITS + UD_HOST_BITS); \
}

#define INCREMENT_SEQ_NUM(_seqnum) { \
    *(_seqnum) = ((*_seqnum) + 1) % MAX_SEQ_NUM; \
    if(0 == *(_seqnum)) { *(_seqnum) = 1; } \
}

#define GET_NEXT_PKT(_dest) mvdev.connections[_dest].seqnum_next_tosend

#define INC_NEXT_PKT(_dest) { \
    mvdev.connections[_dest].seqnum_next_tosend = (mvdev.connections[_dest].seqnum_next_tosend + 1) % MAX_SEQ_NUM; \
    if(0 == mvdev.connections[_dest].seqnum_next_tosend) { \
        mvdev.connections[_dest].seqnum_next_tosend = 1; \
    } \
}

#define INCL_BETWEEN(_val, _start, _end) \
    (((_start > _end) && (_val >= _start || _val <= _end)) || \
    ((_end > _start) && (_val >= _start && _val <= _end)) || ((_end == _start) && (_end == _val)))


#define PEEK_RECV_WIN(_c)  (_c)->recv_window_head->seqnum

typedef struct {
    struct ibv_cq      *send_cq;
    struct ibv_cq      *recv_cq;
    struct ibv_srq     *srq;
    struct ibv_pd      *pd;
    struct ibv_qp_cap  cap;
    uint32_t           sq_psn;
} mv_qp_setup_information;

typedef struct poolptr {
    void *next;
    void *prev;
} poolptr;

typedef struct {
    struct ibv_device * device;
    struct ibv_device_attr device_attr;
    struct ibv_pd * pd;
    struct ibv_context * context;
    struct ibv_port_attr * port_attr;
    struct ibv_port_attr * default_port_attr;
    pthread_t async_thread;
} mv_hca;

/* A new allocation of this structure is required for 
 * each QP that is created... so we may require additional
 * QPs only for some connections and not others
 */
typedef struct _mv_qp {
    struct ibv_qp *qp;
    uint16_t last_qp;
    int send_wqes_avail;
    int send_wqes_total;

    uint16_t unsignaled_count;


    uint16_t max_inline;
    uint16_t max_send_size;

    mv_sdescriptor *ext_sendq_head;
    mv_sdescriptor *ext_sendq_tail;
    uint16_t ext_sendq_size;

    mv_hca * hca;
    struct ibv_port_attr * port;

    struct _mv_qp * next;
} mv_qp;

typedef struct _mvdev_channel_rc {
    mv_qp qp;

    uint16_t send_credits_remaining;
    uint16_t recv_credits_remaining;
    uint16_t recv_credits_toreturn;
    uint8_t  srq;

} mvdev_channel_rc;


typedef struct _mvdev_connection_t {
    int global_rank;
    int rc_enabled;
    int rcfp_enabled;

    uint64_t total_messages;
    uint64_t total_volume;

    int check_for_acks;

    /* this needs to be an array with a user-defined maximum */
    mvdev_channel_rc ** channel_rc;

    /* These are global sequence numbers that span
     * all channels. Can allow reliability with CRC 
     * checks, etc. Also allowes re-ordering between
     * channels.
     */
    packet_sequence_t seqnum_next_tosend;
    packet_sequence_t seqnum_next_toack;

    mv_rbuf *           recv_window_head;
    mv_rbuf *           recv_window_tail;
    int                 recv_window_size;
    packet_sequence_t   recv_window_start_seq;

    mv_sbuf *           send_window_head;
    mv_sbuf *           send_window_tail;
    int                 send_window_size;
    int                 send_window_segments;

    mv_sbuf *           ext_window_head;
    mv_sbuf *           ext_window_tail;

    int                 ud_remote_credit;
    int                 r3_segments;
    int blocked_count;

    MPIR_SHANDLE *shandle_head; /* "queue" of send handles to process */
    MPIR_SHANDLE *shandle_tail;
    MPIR_RHANDLE *rhandle;      /* current eager receive "in progress" */

    struct _mvdev_connection_t *nextflow;

    int inflow;

    struct ibv_ah ** data_ud_ah;
    int last_ah;
} mvdev_connection_t;

typedef struct {
    uint32_t seqnum;
    int32_t  associated_qpn;
    int32_t  associated_rank;
    uint16_t index;

    struct ibv_qp * ud_qp;
    struct ibv_cq * ud_cq;

    mv_hca * hca;

    poolptr ptr;
} mv_qp_pool_entry;

typedef struct {
    void  (*post_ud_send)(mvdev_connection_t * c, mv_sbuf * v, int total_len, mv_qp *send_qp, int signalled, int lid);
} mv_func;


typedef struct {
    /* these are only for the basic UD QP that gets
     * setup at the beginning
     */
    uint32_t *lids;
    uint32_t *qpns;

    mv_hca * hca;
    mv_hca * default_hca;
    int num_hcas;

    mvdev_connection_t *connections;
    uint8_t *connection_ack_required;

    int num_ud_qps;
    int last_qp;
    mv_qp * ud_qp;

    int num_cqs;
    struct ibv_cq ** cq;

    uint32_t posted_bufs;
    
    mv_sbuf * unack_queue_head;
    mv_sbuf * unack_queue_tail;

    mv_qp * rndv_qp;
    struct ibv_cq ** rndv_cq;

    /* more general things */
    int np;
    int me;
    int global_id;
    char *my_name;

    uint64_t total_volume;
    uint64_t total_messages;

    /* qp pool */
    mv_qp_pool_entry *rndv_pool_qps;
    mv_qp_pool_entry *rndv_pool_qps_free_head;
    mv_qp_setup_information rndv_si;

    char * grh_buf;
    struct ibv_mr * grh_mr[MAX_NUM_HCAS];

    int *pids;                    /* add for totalview */
    char **processes;
    char execname[256];           /* add for totalview */
} mvdev_info_t;


extern mvdev_info_t mvdev;
extern mv_func mvdev_func;

extern char nullsbuffer, nullrbuffer;

void mvdev_process_send_afterack(mv_sbuf * v);

/* start the threads that provide various services */
void MV_Start_Threads();

struct ibv_mr* register_memory(int hca_num, void *buf, int len);
int deregister_memory(struct ibv_mr *);

#define REQ_TO_ID(h) ((request_id_t)(h))
#define ID_TO_REQ(id)((union MPIR_HANDLE *)(id))


#define SEND_COMPLETE(s) {                                          \
    s->is_complete = 1;                                             \
    if (s->finish != NULL) s->finish(s);                            \
    if (s->dreg_entry != NULL) {                                    \
        dreg_unregister(s->dreg_entry);                             \
    }                                                               \
    s->dreg_entry = NULL;                                           \
    if (s->ref_count == 0) {                                        \
        switch (s->handle_type) {                                   \
            case MPIR_SEND:                                         \
            {                                                       \
                MPID_SendFree(s);                                   \
                    break;                                          \
            }                                                       \
            case MPIR_PERSISTENT_SEND:                              \
            {                                                       \
                MPID_PSendFree(s);                                  \
                break;                                              \
            }                                                       \
            default:                                                \
                error_abort_all(GEN_EXIT_ERR, "SEND_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}

#define RECV_COMPLETE(r) {                                          \
    r->is_complete = 1;                                             \
    r->s.MPI_ERROR = MPI_SUCCESS;                                   \
    r->s.count = r->len;                                            \
    if (r->finish != NULL) {                                        \
        r->finish(r);                                               \
    }                                                               \
    if (r->dreg_entry != NULL) {                                    \
        dreg_unregister(r->dreg_entry);                             \
    }                                                               \
    r->dreg_entry = NULL;                                           \
    if (r->ref_count == 0) {                                        \
        switch (r->handle_type) {                                   \
            case MPIR_RECV:                                         \
            {                                                       \
                MPID_RecvFree(r);                                   \
                break;                                              \
            }                                                       \
            case MPIR_PERSISTENT_RECV:                              \
            {                                                       \
                MPID_PRecvFree(r);                                  \
                    break;                                          \
            }                                                       \
            default:                                                \
                error_abort_all(GEN_EXIT_ERR, "RECV_COMPLETE invalid type\n");    \
        }                                                           \
    }                                                               \
}

#define MPID_CommInit(oldcomm,newcomm) MPI_SUCCESS
#define MPID_CommFree(oldcomm) MPI_SUCCESS


#define IBV_POST_SR(v, c, err_string) {                             \
    {                                                               \
        struct ibv_wr *bad_wr;                                      \
        if(ibv_post_send(c->vi, &(v->desc.u.sr), &bad_wr)) {        \
            error_abort_all(IBV_RETURN_ERR, err_string);            \
        }                                                           \
    }                                                               \
}

/* See if the SRQ is full or if we need to post more buffers
 * to it
 */

#define CHECK_RECV_BUFFERS() {                                      \
    if(mvdev.posted_bufs <= mvparams.credit_preserve) {             \
        mvdev.posted_bufs +=                                        \
            mvdev_post_srq_buffers(                                 \
                 mvparams.recvq_size - mvdev.posted_bufs);          \
    }                                                               \
}
#define CHECK_RECV_BUFFERS_LAZY() {                                 \
    if(mvdev.posted_bufs <= mvparams.credit_preserve_lazy) {        \
        mvdev.posted_bufs +=                                        \
            mvdev_post_srq_buffers(                                 \
                 mvparams.recvq_size - mvdev.posted_bufs);          \
    }                                                               \
}


#define RENDEZVOUS_IN_PROGRESS(c, s) {                              \
    if (NULL == c->shandle_tail) {                                  \
        c->shandle_head = s;                                        \
    } else {                                                        \
        c->shandle_tail->nexthandle = s;                            \
    }                                                               \
    c->shandle_tail = s;                                            \
    s->nexthandle = NULL;                                           \
}

#define RENDEZVOUS_DONE(c) {                                        \
    MPIR_SHANDLE *_s = c->shandle_head; \
    c->shandle_head = c->shandle_head->nexthandle;                  \
    _s->nexthandle = NULL; \
        if (NULL == c->shandle_head) {                              \
            c->shandle_tail = NULL;                                 \
        }                                                           \
}


void mvdev_eager_pull(MPIR_RHANDLE * rhandle);
void mvdev_rendezvous_reply(MPIR_RHANDLE * rhandle);
void mvdev_recv_r3(MPIR_RHANDLE * rhandle);
int mvdev_post_srq_buffers(int num_bufs);
void mvdev_copy_unexpected_handle_to_user_handle(
        MPIR_RHANDLE * rhandle,
        MPIR_RHANDLE * unexpected,
        int *error_code);

void flush_all_messages();
void async_thread(void *context);

void mvdev_incoming_rendezvous_reply(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_rendezvous_reply * header);
void mvdev_incoming_rendezvous_start(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_rendezvous_start * header);

void mvdev_incoming_ud_zcopy_finish(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_ud_zcopy_finish * h);
void mvdev_incoming_ud_zcopy_ack(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_ud_zcopy_ack * h);
void mvdev_process_recv_inorder(mvdev_connection_t * c, mv_rbuf * v);

void mvdev_post_channel_send_qp_aux_headers(
                mvdev_connection_t * c, mv_sbuf * v, int total_len, mv_qp *send_qp, int signalled, int lid);
void mvdev_post_channel_send_qp_aux(
                mvdev_connection_t * c, mv_sbuf * v, int total_len, mv_qp *send_qp, int signalled, int lid);

void mvdev_incoming_ud_zcopy_ack(mv_rbuf * v, mvdev_connection_t * c, mvdev_packet_ud_zcopy_ack * h);

void mvdev_rendezvous_push_zcopy(MPIR_SHANDLE * s, mvdev_connection_t *c);

extern mvdev_connection_t *flowlist;

#define PUSH_FLOWLIST(c) {                                          \
    if (0 == c->inflow) {                                           \
        c->inflow = 1;                                              \
        c->nextflow = flowlist;                                     \
        flowlist = c;                                               \
    }                                                               \
}

#define POP_FLOWLIST() {                                            \
    if (flowlist != NULL) {                                         \
        mvdev_connection_t *_c;                                    \
        _c = flowlist;                                              \
        flowlist = _c->nextflow;                                    \
        _c->inflow = 0;                                             \
        _c->nextflow = NULL;                                        \
    }                                                               \
}


void process_flowlist(void);

/* ----------------------------------- */


int MV_Setup_QPs();
void MV_Generate_Address_Handles();
int MV_UD_Initialize_Connection(mvdev_connection_t * c);
void MV_Transition_UD_QP(mv_qp_setup_information *si, struct ibv_qp * qp);

void mvdev_resend(mv_sbuf *v);
void mvdev_recv_ud_zcopy(MPIR_RHANDLE * rhandle);

static inline int mvdev_eager_ok(int len, mvdev_connection_t * c)
{
        /* for now eager is always ok... */
        return 1;
}

int MV_Setup_Rndv_QPs();

/* here is where we'll need to add a lot more logic. By giving the buffer
 * back we are going to determine the protocol and transport to be used
 */

static inline mv_sbuf *get_sbuf(mvdev_connection_t *c, int length_requested)  {
    mv_sbuf *v;

    /* keep some statistics */

    mvdev.total_volume += length_requested;
    mvdev.total_messages++;

    c->total_volume += length_requested;
    c->total_messages++;

    /* we are going over UD */
    length_requested += 40;

    if(mvparams.use_headers) {
        v = get_mv_sbuf(MIN(1900 + sizeof(mvdev_packet_mheader), length_requested));
    } else {
        v = get_mv_sbuf(length_requested);
    }

    v->transport = MVDEV_TRANSPORT_UD;

    return v;
}


#endif /* _MV_PRIV_H */
